import pytest
import kre.tools._tools as tools

### Test hidden functions
def test_get_split_values():
    assert tools._get_split_values("흙") == (18,18,9)
    assert tools._get_split_values("다") == (3,0,0)
    with pytest.raises(ValueError) as e_info:
        x = tools._get_split_values("ㅋ")

def test_get_letters():
    assert tools._get_letters((18,18,9)) == ('ㅎ','ㅡ','ㄺ')
    assert tools._get_letters((3,0,0)) == ('ㄷ','ㅏ','')

def test_get_combined_value():
    assert tools._get_combined_value('ㄷ','ㅏ','ㄺ') == 45805
    assert tools._get_combined_value('ㄷ','ㅏ', '') == 45796
    assert tools._get_combined_value('ㄷ','ㅏ') == 45796

def test_values_to_letter():
    assert tools._values_to_letter(('ㅇ','ㅘ','ㅇ')) == '왕'
    assert tools._values_to_letter(['ㅇ','ㅘ','ㅇ']) == '왕'
    assert tools._values_to_letter(('ㅇ','ㅘ','')) == '와'
    assert tools._values_to_letter(('ㅇ','ㅘ')) == '와'

def test_combine_value_list():
    assert tools._combine_value_list((18,18,9)) == '흙'
    assert tools._combine_value_list((3,0,0)) == '다'

def test_split_coda():
    assert tools.split_coda('ㄺ') == 'ㄹㄱ'
    assert tools.split_coda('닭') == '닭'

def test_combine_coda():
    assert tools._combine_coda('ㄹㄱ') == 'ㄺ'
    assert tools._combine_coda('ㄹㅓ') == 'ㄹㅓ'
    assert tools._combine_coda('ㅎㅡㄹㄱ') == 'ㅎㅡㄹㄱ'

### Test external functions
def test_combine():
    assert tools.combine('ㄷㅏㄺ') == '닭'
    assert tools.combine('ㄷㅏㄹㄱ') == '닭'
    assert tools.combine(('ㄷ', 'ㅏ', 'ㄺ')) == '닭' 
    assert tools.combine(('ㄷ', 'ㅏ', 'ㄹㄱ')) == '닭' 
    assert tools.combine(('ㄷ', 'ㅏ', 'ㄹ','ㄱ')) == '닭' 

def test_split():
    assert tools.split('닭', fill_finals = True) == ['ㄷ', 'ㅏ', 'ㄺ']
    assert tools.split('닭', fill_finals = False) == ['ㄷ', 'ㅏ', 'ㄺ']
    assert tools.split('다', fill_finals = False) == ['ㄷ', 'ㅏ']
    assert tools.split('다', fill_finals = True) == ['ㄷ', 'ㅏ', '']
    assert tools.split('닭', split_codas = True) == ['ㄷ', 'ㅏ', 'ㄹㄱ']
    assert tools.split('닭', fill_finals = True, split_codas = True) == ['ㄷ', 'ㅏ', 'ㄹㄱ']

def test_isSyllable():
    assert tools.isSyllable('닭') == True
    assert tools.isSyllable('w') == False
    assert tools.isSyllable('ㅈ') == False
    assert tools.isSyllable('달걀') == False

def test_isComplexCoda():
    assert tools.isComplexCoda('ㄹ') == False
    assert tools.isComplexCoda('ㄺ') == True
    assert tools.isComplexCoda('ㄹㄱ') == False
    assert tools.isComplexCoda('ㄺ다') == False
    
def test_isComplexOnset():
    assert tools.isComplexOnset('ㄴ') == False
    assert tools.isComplexOnset('ᄓ') == True
    assert tools.isComplexOnset('ㄴㄱ') == False

def test_isJamo():
    assert tools.isJamo('ㄱ') == True
    assert tools.isJamo('ㅣ') == True
    assert tools.isJamo('ㄺ') == True
    assert tools.isJamo('ㄹㄱ') == False

def test_isLetter():
    assert tools.isLetter('ㄱ') == True
    assert tools.isLetter('ㅣ') == True
    assert tools.isLetter('ㄺ') == False
    assert tools.isLetter('ㄺ', include_complex=True) == True
    assert tools.isLetter('ㄹㄱ') == False

def test_isHangul():
    assert tools.isHangul('ㄱ') == True
    assert tools.isHangul('ㄺ') == True
    assert tools.isHangul('닭') == True
    assert tools.isHangul('a') == False
    assert tools.isHangul('.') == False

def test_hasHangul():
    assert tools.hasHangul('ㄱ') == True
    assert tools.hasHangul('ㄺ') == True
    assert tools.hasHangul('닭') == True
    assert tools.hasHangul('a') == False
    assert tools.hasHangul('.') == False
    assert tools.hasHangul('A sentence with 한글.') == True



#KO: Basic Tools for Working with Korean Text
#Version 0.5.2
#Author: Darrell Larsen
#
#Distributed under GNU General Public License v3.0
from collections.abc import Sequence
from ._constants import (_COMBINED_FINALS, _FINALS, _INITIALS, _MEDIALS,
        _SPLIT_FINALS)

"""Section: Conversion between Korean syllables (as Unicode object 
strings) and their component sounds.

For simple conversion of a Korean syllable to its ordinal value, or 
conversion from an ordinal value to a Korean syllable, use Python's
built-in ord() and chr() functions.
"""

# TODO: implement exceptions throughout

def _get_split_values(char) -> tuple[int, int, int]:
    """
    Convert a one-character Korean syllable 3-tuple of dictionary
    look-up values

    Args:
        char (str): a one-character Korean syllable input

    Returns:
        tuple(int, int, int): tuple of look-up values of the initial,
            medial, and final letters of the syllable

    Example:
        '닭' --> (3, 0, 9)
    """

    if not isSyllable(char):
        raise ValueError(
                f"Korean syllable character expected, got '{char}'")

    # Get the Unicode value of the input string
    value_char = ord(char)

    # Convert value_a into the look-up values of the letters it 
    # represents
    fin = ((value_char-44032) % 588) % 28
    mid = (((value_char - 44032) % 588) - fin) / 28
    init = ((value_char -44032) - mid*28 - fin) / 588

    return (int(init), int(mid), int(fin))

def _get_letters(value_set: Sequence[int, int, int]) -> tuple[str,str,str]:
    """
    Convert a sequence of look-up values into a one-character syllable

    Args:
        value_set -- a list or tuple containing look-up values to the 
            initial, medial and final Korean letters

    Returns:
        tuple(str, str, str): a 3-tuple of Korean letters
    
    Example:
        (3, 0, 9) --> ('ㄷ', 'ㅏ', 'ㄺ')
    """

    for key in _INITIALS:
        if _INITIALS[key][0] == value_set[0]:
            init = key
    for key in _MEDIALS:
        if _MEDIALS[key][0] == value_set[1]:
            mid = key
    for key in _FINALS:
        if _FINALS[key][0] == value_set[2]:
            fin = key
    return (init, mid, fin)

def _get_combined_value(
        initial: str, medial: str, final: str = None) -> int:
    """
    Convert a sequence of Korean letters into the Unicode value of the
    syllable they constitute

    Args:
        initial (str): an initial Korean letter
        medial (str): a medial Korean letter
        final (str): a final Korean letter (default is None)

    Returns:
        int: Unicode value of the syllable comprised of init, med, fin

    Example:
        'ㄷ', 'ㅏ', 'ㄺ' --> 45805 
    """
    output = 44032 + _INITIALS[initial][0]*588 + _MEDIALS[medial][0]*28
    
    if final:
        output += _FINALS[final][0]

    return output

def _values_to_letter(char_seq: Sequence[str, str, (str)]) -> str:
    """
    Convert a sequence of Korean letters into their one-character
    syllable representation

    Arguments:
        char_seq (list): a sequence of two or three Korean letters

    Returns:
        str: a one-character syllable

    Example:
        ('ㄷ', 'ㅏ', 'ㄺ') --> '닭' 
    """
    if len(char_seq) == 2:
        char_seq = list(char_seq)[:] + ['']

    return chr(_get_combined_value(char_seq[0],char_seq[1],char_seq[2]))
   
def _combine_value_list(int_seq: Sequence[int,int,int]) -> str:
    """
    Convert a list of look-up values into a one-character syllable

    Args:
        int_seq (list): a sequence of look-up values

    Returns:
        str: one-character syllable

    Example:
        (3, 0, 9) --> '닭'
    """

    return _values_to_letter(_get_letters(int_seq))

def split_coda(final: str) -> str:
    """
    Splits a complex coda/final into separate letters.

    Args:
        final: a Korean letter

    Returns:
        a string containing the split letters or the input if not a 
        complex coda
    """
    if final in _COMBINED_FINALS.keys():
        final = _COMBINED_FINALS[final]

    return final

def _combine_coda(final: str) -> str:
    """
    Combines letters of a complex coda/final into a single character.

    Args:
        final (str): a Korean letter

    Returns:
        the character represent the complex coda/final or the input if
        not a complex coda
    """
    if final in _SPLIT_FINALS.keys():
        final = _SPLIT_FINALS[final]

    return final


#### External Functions ####

def combine(*args: str or Sequence[str]) -> str:
    """
    Combines 2-4 Korean letters into a single syllable

    Args:
        str or letter-separated (list or tuple)

    Returns:
        str: a one-character syllable

    Example:
        'ㄷㅏㄺ' --> '닭'
        'ㄷㅏㄹㄱ' --> '닭'
        ('ㄷ', 'ㅏ', 'ㄺ') --> '닭' 
        ('ㄷ', 'ㅏ', 'ㄹㄱ') --> '닭' 
        ('ㄷ', 'ㅏ', 'ㄹ','ㄱ') --> '닭' 

    """

    # If length one, should be str, list, or tuple
    # Sets args = list, tuple, or list(str)
    if len(args) == 1:
        args = args[0]
        if type(args) is str:
            args = list(args)

    if 1 < len(args) < 5:
        # combine two-letter complex finals, if any
        final = _combine_coda(''.join(args[2:]))

        # construct and return syllable
        return chr(_get_combined_value(args[0], args[1], final))

    elif len(args) > 4:
        raise Exception("Too many characters")
    else:
        raise Exception("Insufficient characters")
        
def syllabify(text, linearize_first=True) -> str:
    """
    Combines a string of Korean letters into syllables.

    Args:
        text (str): a string which presumably contains at least some
            Korean letters
        linearize_first (bool): if True, any Korean syllables or complex 
            coda characters in the input will be linearized first, then
            syllabified 

            ex. True: 가ㄹ -> ㄱㅏㄹ -> 갈
            ex. False: 가ㄹ -> 가ㄹ -> 가ㄹ

    Returns:
        str: a string of in which any linear sequences of possible 
        Korean syllables have been replaced with Korean syllables
    """

    buffer = ''
    output = ''

    if linearize_first == True:
        text = linearize(text, split_codas=True)

    for char in text:
        if not isLetter(char, include_complex=True):
            if len(buffer) > 1:
                output += combine(buffer)
            else:
                output += buffer
            output += char
            buffer = ''
            continue

        elif not buffer:
            # only accept an onset letter
            if char not in _INITIALS.keys():
                output += char
            else:
                buffer = char
            continue

        elif len(buffer) == 1:
            if char in _MEDIALS.keys():
                buffer += char
            else:
                output += buffer
                if char in _INITIALS.keys():
                    buffer = char
                else:
                    buffer = ''
            continue

        elif len(buffer) == 2:
            if char in _COMBINED_FINALS.keys():
                output += combine(buffer + char)
                buffer = ''
            elif char in _FINALS.keys():
                buffer += char
            else:
                output += combine(buffer)
                if char in _INITIALS.keys():
                    buffer = char
                else:
                    output += char
                    buffer = ''
            continue

        elif len(buffer) == 3: 
            if (buffer[-1] + char) in _SPLIT_FINALS.keys():
                buffer += char
            elif char in _INITIALS.keys():
                output += combine(buffer)
                buffer = char
            elif char in _MEDIALS.keys():
                output += combine(buffer[:-1])
                buffer = buffer[-1] + char
            else:
                output += combine(buffer) + char
                buffer = ''
            continue
        
        elif len(buffer) == 4:
            if char in _INITIALS.keys():
                output += combine(buffer)
                buffer = char
            elif char in _MEDIALS.keys():
                output += combine(buffer[:-1])
                buffer = buffer[-1] + char
            else:
                output += combine(buffer) + char
                buffer = ''
            continue

        else:
            raise Exception("This shouldn't happen")

    if len(buffer) > 1:
        output += combine(buffer)
    else:
        output += buffer

    return output

def split(char, 
        fill_finals=False, split_codas=False) -> list[str, str, (str)]:
    """
    Converts a Korean syllable character into a sequence of letters

    Args:
        char (str): a Korean syllable character
        fill_finals (bool): return an empty string in finals position
            if no final is provided in the input (default is False)

    Returns:
        list(str, str(, str)): a list of Korean letters, optionally
            includes an empty string to occupy empty final letter
            position

    Examples:
        '닭' --> ['ㄷ', 'ㅏ', 'ㄺ']     (fill_finals=True or False)
        '다' --> ['ㄷ', 'ㅏ']           (fill_finals=False)
        '다' --> ['ㄷ', 'ㅏ', '']       (fill_finals=True)
        '닭' --> ['ㄷ', 'ㅏ', 'ㄹㄱ']   (split_codas=True
    """

    output = list(_get_letters(_get_split_values(char)))
    if split_codas:
        output[-1] = split_coda(output[-1])

    # Remove empty strings in finals position if fill_finals is False
    if output[-1] == '' and not fill_finals:
        output = output[:-1]

    return output

def linearize(text, split_codas=True) -> str:
    output = ''
    for char in text:
        if isSyllable(char):
            output += ''.join(split(char, fill_finals=False, 
                    split_codas=split_codas))
        elif char in _COMBINED_FINALS.keys() and split_codas == True:
            output += ''.join(split_coda(char))
        else:
            output += char
    return output

"""Section: Tests for Korean input."""

def isSyllable(text) -> bool:
    """
    Test whether an input string is a Korean syllable or not

    Args:
        text (str): the string to test

    Returns:
        bool: True if the input string is of length one and is a Korean
            syllable (i.e., is located within the Unicode range
            specified below). False otherwise.

    TODO: implement other Unicode ranges containing Korean syllables;
    currently only returns true for modern Korean syllables using the
    default unicode space
    """

    if len(text) == 1 and 44032 <= ord(text) <= 55203:
        return True
    else:
        return False

def isComplexCoda(char) -> bool:
    """
    Test whether the character is a coda character containing two 
    letters (i.e., a consonant cluser).

    Ssangjaeum (doubled consonants) are *not* treated as consonant
    clusters by this algorithm.

    Args:
        char (char): a single character

    Returns:
        bool: True if the character represents two adjacent coda 
        consonants, False otherwise
    """

    coda_ccs = []

    # add coda clusters in U1100 set (4520-4607)
    coda_ccs += [4522, 4524, 4525, 4528, 4529, 4530, 4531, 4532, 4533,
            4534, 4537, 4547, 4748, 4549, 4550, 4551, 4552, 4553, 4554,
            4555, 4556, 4557, 4558, 4559, 4560, 4561, 4562, 4563, 4564,
            4565, 4566, 4567, 4568, 4569, 4570, 4571, 4572, 4573, 4574,
            4575, 4576, 4577, 4579, 4580, 4581, 4583, 4584, 4585, 4586,
            4588, 4589, 4590, 4591, 4593, 4594, 4595, 4597, 4598, 4599,
            4600, 4602, 4603, 4604, 4605, 4606]

    #add coda clusters in U3130 set (12593-12686)
    coda_ccs += [12595, 12597, 12598, 12602, 12603, 12604, 12605,12606,
            12607,12608,12612, 12647, 12648, 12649, 12650, 12651, 
            12652, 12653, 12655, 12656]

    # add coda clusters in UD7B0 set (55216-55291)
    # note that this set excludes initial consonants (see corresponding
    # set UA960)
    coda_ccs += [55243, 55243, 55244, 55246, 55247, 55248, 55249, 55250,
            55251, 55252, 55253, 55254, 55255, 55256, 55257, 55258, 
            55259, 55260, 55262, 55263, 55265, 55266, 55267, 55268, 
            55269, 55271, 55272, 55273, 55274, 55275, 55276, 55277, 
            55278, 55279, 55280, 55281, 55282, 55283, 55284, 55285, 
            55286, 55287, 55288, 55290]


    if len(char) == 1 and ord(char) in coda_ccs:
        return True
    else:
        return False

def isComplexOnset(char) -> bool:
    """
    Test whether the character is an onset character containing two 
    letters (i.e., a consonant cluser). Note that this does not occur in
    modern Korean writing.
    
    Ssangjaeum (doubled consonants) are *not* treated as consonant
    clusters by this algorithm.


    Args:
        char (char): a single character

    Returns:
        bool: True if the character represents two adjacent onset 
        consonants, False otherwise
    """

    onset_ccs = []

    #check in U1100 set (4371-4446)
    onset_ccs += [4371, 4373, 4374, 4375, 4376, 4378, 4380, 4382, 4383,
            4384, 4385, 4386, 4387, 4388, 4389, 4390, 4391, 4392, 4393,
            4394, 4397, 4398, 4399, 4400, 4401, 4402, 4403, 4404, 4405,
            4406, 4407, 4408, 4409, 4410, 4411, 4417, 4418, 4419, 4420,
            4421, 4422, 4424, 4425, 4426, 4427, 4429, 4434, 4435, 4438,
            4442, 4443, 4444, 4445, 4446]

    #check in U3130
    onset_ccs += [12646, 12654, 12658, 12659, 12660, 12661, 12662, 
            12663, 12666, 12667, 12668, 12669, 12670, 12674, 12675]

    # add onset clusters in UA960 set (43360-43388)
    # note that set UA960 exclusively contains old initial consonants
    onset_ccs += [x for x in range(43360, 44385)]
    onset_ccs += [43386, 43387]

    if len(char) == 1 and ord(char) in onset_ccs:
        return True
    else:
        return False

def isJamo(char) -> bool:
    """
    Test whether the character is Jamo (Korean letter, including
    complex letters) rather than a syllable.

    Args:
        char (char)

    Returns:
        bool: True if char is a Korean Jamo, otherwise False
    """

    return isLetter(char, include_complex=True)

def isLetter(char, include_complex=False) -> bool:
    """
    Test whether the character is letter rather than a syllable. Onset
    and coda clusters return a value of False. (Use isJamo() to accept
    complex clusters as well.)

    Args:
        char (char)

    Returns:
        bool: True if char is a single Korean letter (optionally
        including clusters), otherwise False

    """

    if len(char) != 1:
        return False

    if include_complex == False:
        if isComplexOnset(char) or isComplexCoda(char):
            return False

    value = ord(char)

    #check in U3130 set (12593-12686), and exclude empty values
    if value >= 12593 and value <= 12686 and value != 12644:
        return True
	
    #check in U1100 set (4352-4469), and exclude empty values
    elif value >= 4352 and value <= 4607 and value not in [4447, 4448]:
        return True

    #check in UA960 set (43360-43388)
    elif value >= 43360 and value <= 43388:
        return True

    #check in UD7B0 set (55216-55291), and exclude empty values
    elif value >= 55216 and value <= 55291 and (
            value not in [55239, 55240, 55241, 55242]):
        return True
    else:
        return False

def isHangul(char) -> bool:
    """
    Checks whether input is a Korean Hangul (and not Hanja) symbol.

    Args:
        char (char)

    Returns:
        bool: True if char is a Korean Hangul symbol, including
        syllables, individual letters, and clusters. False otherwise.
    """
    if isSyllable(char) or isJamo(char):
        return True
    else:
        return False

def hasHangul(text) -> bool:
    """
    Checks whether the input string contains Korean Hangul (and not Hanja)

    Args:
        text (str)

    Returns:
        bool: True if text contains a Korean Hangul symbol, including
        syllables, individual letters, and clusters. False otherwise.
    """
    for char in text:
        if isHangul(char):
            return True

    return False

#### Romanization

def toYale(text, syllable=None, WO=False, U=False, strict=False) -> str:
    """
    Options: 
    syllable = max (will place a dot between all syllables in a word)
    syllable = min (limits dots to official Yale rules, with a few extra)
    syllable = none (no syllables boundaries marked)
    WO = True (will use <wo> instead of <o> for ㅗ.  Useful for some time periods)
    U = True (will use <u> after labial consonants)
    strict = True (follows Yale Romanization rules as closely as possible, without
    phonological knowledge; currently, this sets WO and U to true, and syllable to min, 
    regardless what the user entered
    """
    
    output = ''
    prev_letters = None
    if strict == True:
        WO = True
        U = True
        syllable = 'min'
   
    for char in text:
        if isHangul(char) and isSyllable(char):
            letters = split(char)
            onset = _INITIALS[letters[0]][1]
            nucleus = _MEDIALS[letters[1]][1]
            coda = _FINALS[letters[2]][1]

            if WO == True and nucleus == 'o':
                nucleus = 'wo'
            if U == True and nucleus == 'wu' and onset in {'p','pp','ph','m'}:
                nucleus = 'u' 

            #syllable = 'max': places dot between all syllables
            if prev_letters != None and syllable == 'max':
                output = output + '.'

            #syllable = 'min': places dot between ambiguous syllables only    
            elif prev_letters != None and syllable == 'min':
                prev_coda = prev_letters[2]
                prev_vowel = prev_letters[1]
            
                if prev_coda == '':
                    #Mark after coda-less syllables when C follows
                    if onset != '':
                        output = output + '.'

                    #Mark between adjacent vowels, when confusion of <w>/<y> 
                    #and <a>, <e>, <o>, <u> may arise
                    elif ((prev_vowel[len(prev_vowel)-1] in {'a','e','o','u'} and \
                            nucleus == 'y') \
                            or (prev_vowel[len(prev_vowel)-1] in {'y', 'w'} and \
                            nucleus in {'a','e','o','u'})):
                        output = output + '.'

                else:        
                    #Mark before onsetless syllables, unless previous coda was 
                    #null or <ng>
                    if prev_coda not in {'','ng'} and onset == '':
                        output = output + '.'
                        

                    #Mark after a consonant (final one in case of cluster) when onset 
                    #consonant is same, for those which can be doubled
                    elif prev_coda[len(prev_coda)-1] == onset and onset in \
                        {'k','p','t','c','s'} and prev_coda not in \
                        {'kk','pp','tt','cc','ss'}:
                        output = output + '.'

                    #Mark after a consonant when onset of following is <h>, when 
                    #ambiguity can arise
                    elif prev_coda[len(prev_coda)-1] in {'k','p','t','c'} and \
                            onset in {'h'}:
                        output = output + '.'

                    #Mark when <k> is followed by <kh>.  Need not do for <p,t,c>, i
                    #as their doubled variants do not occur in coda position    
                    elif prev_coda in {'k'} and onset in {'kh'}:
                        output = output + '.'

            output = (output + onset + nucleus + coda) 
            prev_letters = (onset,nucleus,coda)
 
        else:
            output = output + char
            prev_letters = None
    return output            



#KRE: Support for Korean regular expression searches.
#Author: Darrell Larsen
#
#Distributed under GNU General Public License v3.0

"""
This module is built on top of re.  See re documentation for information
on special characters, etc, not affected by this module.

This module provides support for conducting regular expression searches
over Unicode strings containing Korean text.  This is designed for users
wishing to search for Korean text at the sub-syllable level; those 
wishing to conducted re searches using full Korean syllables can do so 
using the standard re module.  This module breaks Korean syllables up 
into their composite letters in both the regular expression and the text
to be searched prior to conducting the search.  By default, indices are 
then calculated based on the original input text, and a kre.SRE_Match 
object is returned.  kre.KRE_Match objects are designed to behave like
_sre.SRE_Match objects, with a few additional attributes.  As such, the 
output of this module may usually be handled in the same manner as the 
output of the standard re module. 

With the exception of the set of vowels being placed after the set of 
consonants, the precedence of characters is based the order used in 
South Korean dictionaries (see Unicode chart U3130). Older characters 
are placed after all modern consonants and vowels. 

[ㄱ-ㅎ] : matches all Korean consonants and null ㅇ in initial position
[ㅏ-ㅣ] : matches all Korean vowels
[ㄱ-ㅣ] : matches all Korean letters

By default, syllable-final consonant clusters are NOT split (i.e. ㅄ 
is treated as a single character; it is not split into ㅂㅅ), though 
the option to split these up should be implemented at some point. 
Likewise, vowels such as ㅖ and ㅢ are treated as single characters
by default, but researchers may find it help to be able to split them.

By default, kre does NOT track boundaries between syllable characters, 
even if the search string is entered as a syllabary (e.g. 일 rather 
than ㅇㅣㄹ).  Thus, a search for 'ㅣㄹ' will match both '일' and
'이라'.  To track boundaries, set boundaries=True.  The default 
boundary symbol is the semicolon <;> and may be changed using 
delimiter=new_symbol.


TODO:
    Implement exceptions; may want to add kre-specific exceptions

NOTE:
    re.finditer returns class 'callable_iterator', where kre returns
    'list_iterator'. Unclear whether there is any practical difference.
"""

import re

# Make flags from re library accessible 
from re import (A, ASCII, DEBUG, DOTALL, I, IGNORECASE, L, LOCALE, M,
        MULTILINE, T, TEMPLATE, U, UNICODE, VERBOSE, X)
from .tools import _tools
from .tools._constants import _COMBINED_FINALS

_settings = {"boundaries": False,
        "delimiter": ";",
        "syllabify": "minimal",
        "empty_es": True,
        }


### Public interface

def search(pattern, string, flags=0, empty_es=True, **pattern_kwargs):
    return compile(pattern, flags, **pattern_kwargs).search(string, 
            empty_es=empty_es)

def match(pattern, string, flags=0, empty_es=True, **pattern_kwargs):
    return compile(pattern, flags, **pattern_kwargs).match(string, 
            empty_es=empty_es)

def fullmatch(pattern, string, flags=0, empty_es=True, **pattern_kwargs):
    return compile(pattern, flags, **pattern_kwargs).fullmatch(string, 
            empty_es=empty_es)

def sub(pattern, repl, string, count=0, flags=0, empty_es=True, 
        syllabify='minimal', **pattern_kwargs):
    """
    Returns unsubstituted characters in the same format as input (i.e.,
    as syllable characters or individual letters) except as affected by 
    the syllabify options.

    syllabify options: 
        'full' (linearize + syllabify entire string)
        'none' (affected characters are output without syllabification)
        'minimal' (affected characters are syllabified prior to output)
        'extended' (will attempt to combine affected characters with
            preceding/following characters to create syllables)
    """
    return compile(pattern, flags, **pattern_kwargs).sub(repl, string, 
            count=count, empty_es=empty_es, syllabify=syllabify)

def subn(pattern, repl, string, count=0, flags=0, empty_es=True, 
        syllabify='minimal', **pattern_kwargs):
    """
    Similar to sub(), but returns tuple with count as second element.
    """
    return compile(pattern, flags, **pattern_kwargs).subn(repl, string, count=count, 
            empty_es=empty_es, syllabify=syllabify)

def split(pattern, string, maxsplit=0, flags=0, empty_es=True, **pattern_kwargs):
    return compile(pattern, flags, **pattern_kwargs).split(string, maxsplit=maxsplit, 
            empty_es=empty_es,)

def findall(pattern, string, flags=0, empty_es=True, **pattern_kwargs):
    return compile(pattern, flags, **pattern_kwargs).findall(string, 
            empty_es=empty_es,)

def finditer(pattern, string, flags=0, empty_es=True, **pattern_kwargs):
    return compile(pattern, flags, **pattern_kwargs).finditer(string, 
            empty_es=empty_es,)

def compile(pattern, flags=0, **pattern_kwargs):
    return KRE_Pattern(pattern, flags, **pattern_kwargs)

def purge():
    # note that this will purge all regular expression caches, 
    # not just those created through kre
    re.purge()

def escape(pattern):
    return re.escape(pattern)

### Private interface

class KRE_Pattern:
    def __init__(self, pattern, flags, **pattern_kwargs):
        self.pattern = pattern #original Korean, unlinearized
        self.flags = flags
        self.mapping = Mapping(pattern)
        self.linear = self.mapping.linear #linear input to compile
        self.Pattern = re.compile(self.linear, flags) # re.Pattern obj
        self.groups = self.Pattern.groups
        self.boundaries = pattern_kwargs.pop("boundaries",
                _settings["boundaries"])
        self.delimiter = pattern_kwargs.pop("delimiter",
                _settings["delimiter"])

        # Extract from compiled non-linearized string so access format
        # can match input format
        self._re = re.compile(self.pattern)
        self.groupindex = self._re.groupindex

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "kre.compile(%s)" % repr(self.pattern)

    def search(self, string, *args, empty_es=True):
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)
        return self._search(sm, *args, empty_es=empty_es)

    def _search(self, string_mapping, *args, empty_es=True):
        sm = string_mapping
        pos_args, iter_span = self._process_pos_args(sm, *args)
        match_ = self.Pattern.search(sm.linear, *pos_args)
        if match_:
            return KRE_Match(self, sm, match_, *args, empty_es=empty_es)
        else:
            return match_

    def match(self, string, *args, empty_es=True):
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)
        pos_args, iter_span = self._process_pos_args(sm, *args)
       
        for span in iter_span:
            match_ = self.Pattern.match(sm.linear, *span)
            if match_:
                return KRE_Match(self, sm, match_, *args, 
                        empty_es=empty_es)
        else:
            return match_

    def fullmatch(self, string, *args, empty_es=True):
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)
        pos_args, iter_span = self._process_pos_args(sm, *args)

        for span in iter_span:
            match_ = self.Pattern.fullmatch(sm.linear, *span)
            if match_:
                return KRE_Match(self, sm, match_, *args, 
                        empty_es=empty_es)
        else:
            return match_

    def sub(self, repl, string, count=0, empty_es=True, 
            syllabify='minimal'):
        """
        Returns unsubstituted characters in the same format as input (i.e.,
        as syllable characters or individual letters) except as affected by 
        the syllabify options.

        syllabify options: 
            'none' (affected characters are output without syllabification)
            'minimal' (affected characters are syllabified prior to output)
            'extended' (will attempt to combine affected characters with
                the immediately preceding/following characters to create 
                syllables when it minimizes stand-alone Korean letters)
            'full' (linearize + syllabify entire string; deletes
                    boundaries prior to syllabifying string)
        """
        # Linearize string
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)

        return self._sub(repl, sm, count=count, empty_es=empty_es,
                syllabify=syllabify)

    def _sub(self, repl, string_mapping, count=0, empty_es=True, 
            syllabify='minimal'):
        sm = string_mapping
        subs = dict()
        matches = self._finditer(sm)

        def compute_spans():
            i = 0 # number non-overlapping sub spans (no increment for shared syllable)
            for n, match_ in enumerate(matches):
                # limit matches to number indicated by count (0=no limit)
                if 0 < count <= n:
                    break

                lin_span = match_.Match.span()
                del_span = get_corresponding_del_span(lin_span)

                i = add_spans(subs, i, lin_span, del_span)

        def get_corresponding_del_span(lin_span):
            # Case of empty string match at end of string
            if lin_span[0] == len(sm.linear):
                del_span = tuple([sm.lin2del[lin_span[0]-1]+1]*2)

            # Normal case
            else:
                del_span = (sm.lin2del[lin_span[0]],
                    sm.lin2del[lin_span[1]-1]+1)
            return del_span

        def add_spans(subs, i, lin_span, del_span):
            # CASE: multiple subs from same syllable
            if i > 0 and subs[i-1]['del_span'][1] > del_span[0]:
                prev = subs[i-1]
                prev['num_subs'] += 1

                # update endpos
                prev['del_span'] = (prev['del_span'][0], del_span[1])
                prev['lin_span'] = (prev['lin_span'][0], lin_span[1])

            # DEFAULT CASE:
            else:
                subs[i] = dict()
                sub = subs[i]
                sub['num_subs'] = 1
                sub['del_span'] = del_span
                sub['lin_span'] = lin_span
                i += 1
            return i

        def get_outlying_chars():
            # Keep track of extra letters in the subbed syllables which
            # preceded/followed the actual substitution
            for sub in subs.values():
                sub_start, sub_end = sub['lin_span']

                # Case: string-final empty string match
                if sub_start == len(sm.linear):
                    syl_start = len(sm.linear)
                # Normal case
                else:
                    syl_start = sm._get_syl_start(sub_start)
                syl_end = sm._get_syl_end(sub_end-1)

                pre_sub_letters = sm.linear[syl_start:sub_start]
                post_sub_letters = sm.linear[sub_end:syl_end]

                sub['extra_letters'] = (pre_sub_letters,post_sub_letters)
           
        def extract_unchanged_spans():
            # Extract the text from the unchanged indices so we can return them
            # without change. (If we use tools.syllabify to reconstruct the entire
            # string, inputs like 가ㅎ (linearized to ㄱㅏㅎ) would be returned 
            # as 갛 even if they weren't matches for substitution. We want to 
            # avoid this.) 
            unchanged_text = []
            for n in range(len(subs) + 1):
                start = 0 if n == 0 else subs[n-1]['del_span'][1]
                end = len(sm.delimited) if n == len(subs) else subs[n]['del_span'][0]
                unchanged_text.append(sm.delimited[start:end])
            return unchanged_text

        def make_substitutions():
            # Carry out substitutions one by one* to identify the indices of each 
            # changed section. *Multiple subs affecting same syllable are
            # carried out at same time.
            extra = 0 # Tracks added/substracted number of letters after sub
            prev_string = sm.linear
            num_subs = 0 # For carrying out subs incrementally
            for sub in subs.values():
                # Carry out next substitution(s)
                num_subs = num_subs + sub['num_subs']
                subbed_string = self.Pattern.sub(repl, sm.linear,
                        count=num_subs)

                # Calculate the start and end indices of the inserted substitution
                sub_start = sub['lin_span'][0] + extra
                extra += len(subbed_string) - len(prev_string)
                sub_end = sub['lin_span'][1] + extra

                # Combine the substitution(s) with the extra letters from
                # affected syllables
                syl_text = sub['extra_letters'][0]
                syl_text += subbed_string[sub_start:sub_end]
                syl_text += sub['extra_letters'][1]
                sub['subbed_syl'] = syl_text

                prev_string = subbed_string

        def reconstruct():
            # Attempt to construct Hangul characters to the desired degree of 
            # syllabification.
            output = ''
            for n in range(len(unchanged_text)):
                output += unchanged_text[n]
                if n < len(subs):
                    new_text = subs[n]['subbed_syl']
                    if syllabify == 'minimal': 
                        output += _tools.syllabify(new_text)
                    elif syllabify == 'extended':
                        # Remove one character from the unchanged text
                        # that follows the current substitution. The 
                        # character will be in original (nonlinearized)
                        # form
                        if unchanged_text[n+1] != '':
                            post = unchanged_text[n+1][0]
                            unchanged_text[n+1] = unchanged_text[n+1][1:]
                            if _tools.isSyllable(post):
                                post = ''.join(_tools.split(post))
                            new_text += post
                        # Remove the character immediately preceding the
                        # current substitutions. The character will have
                        # been syllabified in the previous loop.
                        if output:
                            pre = output[-1]
                            output = output[:-1]
                            if _tools.isSyllable(pre):
                                pre = ''.join(_tools.split(pre))
                            new_text = pre + new_text
                        output += _tools.syllabify(new_text,
                                linearize_first=True)
                    else:
                        output += new_text

            # Remove the delimiter from the output
            # This should precede 'full' syllabify option
            if sm.boundaries == True:
                output = output.replace(sm.delimiter, '')

            if syllabify == 'full':
                output = _tools.syllabify(Mapping(output).linear)

            if syllabify == 'none':
                pass
            return output

        ### MAIN
        compute_spans()
        get_outlying_chars()
        unchanged_text = extract_unchanged_spans()
        make_substitutions()
        output = reconstruct()

        return output

    def subn(self, repl, string, count=0, empty_es=True, syllabify='minimal'):
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)
        
        # Limit substitutions to max of count if != 0
        res = self._findall(sm, empty_es=empty_es)
        if res:
            sub_count = len(res)
        else:
            sub_count = 0
        if 0 < count < sub_count:
            sub_count = count

        return (self._sub(repl, sm, count=count, empty_es=empty_es, 
            syllabify=syllabify), sub_count)

    def split(self, string, maxsplit=0, empty_es=True):
        raise NotImplementedError 

    def findall(self, string, *args, empty_es=True):
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)
        return self._findall(sm, *args, empty_es=empty_es)

    def _findall(self, string_mapping, *args, empty_es=True):
        matches = self._finditer(string_mapping, *args,
                empty_es=empty_es)
        return [match_.group() for match_ in matches] or None

    def finditer(self, string, *args, empty_es=True):
        sm = Mapping(string, boundaries=self.boundaries, delimiter=self.delimiter)
        return self._finditer(sm, *args, empty_es=empty_es)

    def _finditer(self, string_mapping, *args, empty_es=True):
        sm = string_mapping
        pos_args, _ = self._process_pos_args(sm, *args)

        match_ = self.Pattern.finditer(sm.linear, *pos_args)

        # For all re.Match objects in match_
        match_list = []
        for item in match_:
            cur_match = item
            match_list.append(KRE_Match(self, sm, 
                cur_match, *args, empty_es=empty_es))
        return iter(match_list)

    def _process_pos_args(self, _linear, *args):
        """
        Map the pos and endpos args to their corresponding linear positions.

        When boundaries == True, pos and endpos expand to include boundaries

        Returns:
            tuple: pos and endpos, mapped to position in linear string
            iter_span (tuple of tuples): tuples of indices for `match` 
                and `fullmatch` to iterate through, if any (empty if no
                pos arg was provided)
        """
        output = []
        map_ = _linear.lin2orig
        orig_size = len(_linear.original)
        iterate = True
        iter_span = []

        ### Fill in missing args
        # No pos arg?
        if len(args) == 0:
            iterate = False
            args = [0]
        # No endpos arg? 
        if len(args) == 1:
            args = [args[0], len(map_)]

        ### Limit pos / endpos to string length
        for arg in args:
            if 0 <= arg < orig_size:
                output.append(map_.index(arg))
            # >= length of string?
            elif arg >= orig_size:
                output.append(len(map_)) 
            # Less than 0?
            else:
                output.append(0)
        
        if iterate == True:
            # which indices should be iterated through?
            for n, item in enumerate(map_[output[0]:], output[0]):
                if item == map_[output[0]]:
                    iter_span.append(tuple([n, output[1]]))
                else:
                    break
        else:
            iter_span.append(output)


        # If pos is preceded by a boundary marker, expand to include it
        # No need to do this for endpos
        if output[0] > 0:
            if map_[output[0]-1] == None:
                output[0] -= 1
                if iterate == True:
                    iter_span = [tuple(output)] + iter_span

        return tuple(output), tuple(iter_span)

class Mapping:
    """
    Contains three levels of representation of the input, maps between
    the levels, and methods for navigation, string extraction, and
    mapping representation.

    - levels:
        original: input string without alteration
        delimited: same as original if boundaries == False; otherwise, 
            input string with delimiters placed around Korean characters
            (both syllables and letters)
        linear: linearized form of delimited; i.e., Korean characters
            are all replaced with Korean letter sequences
    - maps: 
        del2orig,
            'This is ;한;글;'
        lin2del, 
        del2lin_span, 
        lin2orig
    - example
        - levels:
            original: 'This is 한글ㅋㅋ'
            delimited: 'This is ;한;글;ㅋ;ㅋ;'
            linear: 'This is ;ㅎㅏㄴ;ㄱㅡㄹ;ㅋ;ㅋ;'
        - maps:
            - Note that maps *to* original level map delimiters to None.
            del2orig: (0,1,2,3,4,5,6,7,None,8,None,9,None,10,None,11,None)
                       T,h,i,s, ,i,s, ,;...,한,;...,글,;...,ㅋ,;...,ㅋ,;...
            lin2del: (0,1,2,3,4,5,6,7,8,9,9,9,10,11,11,11,12,13,14,15,16)
                      T,h,i,s, ,i,s, ,;,ㅎ,ㅏ,ㄴ,;,ㄱ,ㅡ,ㄹ,;ㅋ,;,ㅋ,;
            lin2orig: (0,1,2,3,4,5,6,7,None,8,8,8,None,9,9,9,None,10,None,11,None)
        - span maps:
            - Note that span maps *to* original level map delimiters to
              indices of the form (n, n) (where start=end).
            del2lin_span:
                    ((0,1), #T
                    (1,2), #h
                    (2,3), #i
                    (3,4), #s
                    (4,5), # 
                    (5,6), #i
                    (6,7), #s
                    (7,8), # 
                    (8,9), #;
                    (9,12), #한 -> ㅎㅏㄴ
                    (12,13), #;
                    (13,16), #글 -> ㄱㅡㄹ
                    (16,17), #;
                    (17,18), ㅋ
                    (18,19), #;
                    (19,20), ㅋ
                    (20,21), #;


    Developer notes: 
    - new levels and their mappings should be created simultaneously
    - abbreviations: {delimiter: del; original: orig; linear: lin}
        - use abbreviations exclusively within functions
        - use full name as class attribute
    """
    def __init__(self, string, **kwargs):
        self.boundaries = kwargs.get("boundaries",
        _settings["boundaries"])
        self.delimiter = kwargs.get("delimiter",
        _settings["delimiter"])

        # levels of representation and mappings
        self.original = string
        self.delimited, self.del2orig = self._delimit()
        self.linear, self.lin2del = self._linearize()
        self.lin2orig = tuple(self.del2orig[n] for n in self.lin2del)
        self.del2lin_span = self._get_del2lin_span()
        self.del2orig_span = self._get_del2orig_span()
        self.lin2orig_span = self._get_lin2orig_span()
        self.lin2del_span = self._get_lin2del_span()
        self.orig2del_span = self._get_orig2del_span()
        self.orig2lin_span = self._get_orig2lin_span()

    def validate_delimiter(self) -> None:
        """
        When boundaries == True, checks whether choice of delimiter is
        present in string.
        
        This method should only be called on the 'string' input; it
        should not be called on the 'pattern' input.
        """
        if self.boundaries == True and self.delimiter in self.original:
            raise ValueError('Delimiter must not be present in original string')

    def _delimit(self): # returns (del_str, tuple(del2orig_))
        del_str = ''
        del2orig_ = []
        orig_idx = 0
        just_saw_delimiter = False

        for char_ in self.original:
            if _tools.isHangul(char_):
                
                # add delimiter in front of Korean syllable
                if self.boundaries==True and not just_saw_delimiter:
                    del_str += self.delimiter
                    del2orig_.append(None)

                # add the character
                del_str += char_
                del2orig_.append(orig_idx)

                # add delimiter symbol at end of Korean syllable
                if self.boundaries==True:
                    del_str += self.delimiter
                    del2orig_.append(None)

            else:
                del_str += char_
                del2orig_.append(orig_idx)

            orig_idx += 1
            just_saw_delimiter = (del_str[-1] == self.delimiter)

        return (del_str, tuple(del2orig_))

    def _linearize(self): # returns (lin_str, tuple(lin2del_))
        """
        Linearizes delimited string by splitting up Korean syllables into 
        individual Korean letters.

        Outputs:
            lin_str (str): linearized version of delimited string
            lin2del_ (tuple): index of character positions in delimited string
                ex. given input 한국:
                    lin_str -> ㅎㅏㄴㄱㅜㄱ
                    lin2del_[4] -> 1 (ㅜ in index 1 in input)
        """

        lin_str = ''
        lin2del_ = []
        lin2orig_str = []
        lin_idx = 0

        for char_ in self.delimited:
            # Case: Korean syllable character (e.g., '글')
            if _tools.isSyllable(char_):
                
                # append the linearized string
                for letter in ''.join(_tools.split(char_, split_codas=True)):
                    lin_str += letter
                    lin2del_.append(lin_idx)

                lin2orig_str.append(char_)

            # Case: complex coda character (e.g., 'ㄺ')
            elif char_ in _COMBINED_FINALS.keys():
                # append the linearized string
                for letter in ''.join(_tools.split_coda(char_)):
                    lin_str += letter
                    lin2del_.append(lin_idx)

                lin2orig_str.append(char_)

            else:
                lin_str += char_
                lin2del_.append(lin_idx)
            lin_idx += 1
        return (lin_str, tuple(lin2del_))

    def _get_del2lin_span(self): # returns tuple(span_map)
        span_map = []
        start = 0
        mapped_idx = 0

        for n in range(len(self.lin2del)+1):
            # Case: end of string
            if n == len(self.lin2del):
                span_map.append((start, n))

            # Case: from same syllable as previous
            elif self.lin2del[n] == mapped_idx:
                pass

            # Case: from start of syllable
            else:
                span_map.append((start, n))
                start = n
                mapped_idx += 1

        return tuple(span_map)

    def _get_del2orig_span(self): # returns tuple(span_map)
        span_map = []
        end_idx = 0

        for idx in self.del2orig:
            if idx != None:
                end_idx = idx + 1
                span_map.append((idx, end_idx))
            else:
                span_map.append((end_idx, end_idx))

        return tuple(span_map)

    def _get_lin2orig_span(self):
        span_map = []
        for n in range(len(self.lin2orig)):
            span_map.append(self.del2orig_span[self.lin2del[n]])
        
        return tuple(span_map)

    def _get_lin2orig_span_old(self): # DELETE AFTER TEST
        span_map = []
        end_idx = 0

        for idx in self.lin2orig:
            if idx != None:
                end_idx = idx+1
                span_map.append((idx, end_idx))
            else:
                span_map.append((end_idx, end_idx))

        return tuple(span_map)

    def _get_lin2del_span(self):
        span_map = []
        end_idx = 0

        for idx in self.lin2del:
            if idx != None:
                end_idx = idx+1
                span_map.append((idx, end_idx))
            else:
                span_map.append((end_idx, end_idx))

        return tuple(span_map)

    # Forward maps
    def _get_orig2del_span(self): #Revise - no need for span in this direction
        span_map = []
        map_ = self.del2orig
        _pam = map_[::-1]

        for n in range(len(self.original)):
            start = map_.index(n)
            end = len(map_) - _pam.index(n)
            span_map.append((start, end))

        return tuple(span_map)

    def _get_orig2lin_span(self):
        span_map = []
        map_ = self.lin2orig
        _pam = map_[::-1]

        for n in range(len(self.original)):
            start = map_.index(n)
            end = len(map_) - _pam.index(n)
            span_map.append((start, end))

        return tuple(span_map)

    def _get_syl_span(self, idx):
        return self.del2lin_span[self.lin2del[idx]]

    def _get_syl_start(self, idx):
        return self._get_syl_span(idx)[0]

    def _get_syl_end(self, idx):
        return self._get_syl_span(idx)[1]

    def show_original_alignment(self):
        print('Index\tOriginal\torig2del_span\tdel2orig\tdel2orig_span\tDelimited\tdel2lin_span\tLinear')
        for n, idx in enumerate(self.del2orig):
            orig_idx = idx if idx != None else ''
            orig_str = '' if orig_idx == '' else self.original[orig_idx]
            start, end = self.del2lin_span[n]
            print(f'{n}|{orig_idx}', '\t', orig_str, '\t\t',
                    '\t' if self.del2orig[n] == None else
                            self.orig2del_span[self.del2orig[n]], '\t',
                    self.del2orig[n], '\t\t',
                    self.del2orig_span[n], '\t', self.delimited[n],
                    '\t\t', self.del2lin_span[n], '\t',
                    self.linear[start:end])

    def show_linear_alignment(self):
        print('Index\tLinear\tlin2del\tlin2del_span\tDelimited\tlin2orig\tlin2orig_span\tOriginal')
        for n, span_ in enumerate(self.lin2orig_span):
            d2o_span = self.del2orig_span[self.lin2del[n]]
            print(n, '\t', self.linear[n], '\t', self.lin2del[n], '\t', self.lin2del_span[n],
                    '\t', self.delimited[slice(*self.lin2del_span[n])],
                    '\t\t', self.lin2orig[n], '\t\t', span_,'\t', self.original[slice(*span_)])


def _get_regs(Match, linear_obj):
    # TODO: update doc
    """
    Map the index positions of the match to their corresponding 
    positions in the source text

    Args:
        Match: re.Match object carried out over the linearized text
        linear_obj: Mapping object

    Returns:
        list: a list containing the corresponding span positions in the
            original (non-linearized) text
    """
    regs = []
    l = linear_obj
    for span in Match.regs:
        # (-1, -1) used for groups that did not contibute to the match
        if span == (-1, -1):
            regs.append(span)
            continue

        # Did it match a string-initial empty string?
        elif span == (0, 0):
            regs.append(span)
            continue

        # Did it match a string-final empty string?
        elif span == (len(Match.string), len(Match.string)):
            idx = len(l.original)
            regs.append((idx, idx))
            continue

        else:
            span_start = l.lin2orig_span[span[0]][0]

            # re.MATCH object's span end is index *after* final character,
            # so, we need to subtract one to get the index of the character 
            # to map back to the original, then add one to the result to 
            # get the index after this character
            span_end = l.lin2orig_span[span[1]-1][1]
            regs.append((span_start, span_end))
    
    return tuple(regs)

class KRE_Match:
    """
    The KRE_Match class is intended to mimic the _sre.SRE_Match object,
    but to provide results based on the positions of the Match objects
    in the original input rather than the linearized input which is fed
    to the re module functions. In essence, it converts a _sre.SRE_Match
    object into a new one by replacing the positional arguments of
    letters to their corresponding index positions in the unlinearized
    text.
    A few additional methods are defined to allow the user to obtain data on
    both the original and modified strings created by kre.
    """
    def __init__(self, pattern_obj, string_mapping, Match_obj, *args, 
            empty_es=True):

        self.string_mapping = string_mapping
        # underlying re.Match object 
        # contains same attributes as above but for linearized string
        self.Match = Match_obj
        self.empty_es = empty_es

        self.string = self.string_mapping.original
        self.re = pattern_obj # kre.KRE_Pattern object (kre.compile)
        self._re = self.re.Pattern # re.Pattern object (re.compile)
        self.pos, self.endpos = self._get_pos_args(*args)
        self.regs = _get_regs(Match_obj, self.string_mapping) #tuple
        self.lastindex = Match_obj.lastindex
        self.lastgroup = self._get_lastgroup()
   
        self.linear = string_mapping.linear

    def __repr__(self):
        return "<kre.KRE_Match object; span=%r, match=%r>" % (
                self.span(), self.group(0))

    def __getitem__(self, group):
        return self.group(group)

    def _get_pos_args(self, *args):
        pos_args = [0, len(self.string)] # re defaults
        if args:
            for n, arg in enumerate(args):
                pos_args[n] = arg
        return pos_args

    def expand() -> str:
        """
        expand(template) -> str.
        Return the string obtained by doing backslash substitution
        on the string template, as done by the sub() method.

        NOTE: not well documented. Example use:
        number = re.match(r"(\d+)\.(\d+)", "24.1632")
        print(number.expand(r"Whole: \1 | Fractional: \2"))
        (output) 'Whole: 24 | Fractional: 1632'

        TODO: When implementing, will need to add argument for degrees 
        of syllabification, as with sub() function.
        
        TODO: Will need to treat linearization of pattern different from
        string. Need to maintain identifiers of named groups such as  
        (?P<숫자>\d+) 
        """
        raise NotImplementedError 
    
    def group(self, *args) -> str or tuple:
        """
        group([group1, ...]) -> str or tuple.
        Return subgroup(s) of the match by indices or names.
        For 0 returns the entire match.
        """
        res = []

        if not args:
            args = [0,]
        for arg in args:
            idx = arg
            # convert named groups to group indices
            if isinstance(idx, str):
                idx = self.re.groupindex[idx]
            # non-matching capture group?
            if self.regs[idx] == (-1, -1):
                res.append(None)
            # empty string match? (see discussion on special treatment)
            elif self.empty_es == True and self.Match.group(idx) == '':
                res.append('')
            else:
                res.append(self.string[slice(*self.regs[idx])])

        if len(res) == 1:
            return res[0]
        else:
            return tuple(res)

    def groupdict(self, default=None) -> dict:
        """
        groupdict([default=None]) -> dict.
        Return a dictionary containing all the named subgroups of the 
        match, keyed by the subgroup name. The default argument is used
        for groups that did not participate in the match
        """
        inv_map = {value: key for key, value in
                self.re.groupindex.items()}

        apply_default = lambda d: default if d == None else d

        return {inv_map[n]: apply_default(self.group(n)) for n in inv_map.keys()}

    def groups(self, default=None) -> tuple:
        """
        groups([default=None]) -> tuple.
        Return a tuple containing all the subgroups of the match, from
        1. The default argument is used for groups that did not 
        participate in the match
        """
        g = []
        for n in range(1, len(self.regs)):
            if self.group(n) == None:
                g.append(default)
            else:
                g.append(self.group(n))
        return tuple(g)

    def span(self, *args) -> tuple:
        """
        span([group]) -> tuple.
        For MatchObject m, return the 2-tuple (m.start(group),
        m.end(group)).
        """
        if args:
            idx = args[0]
            if isinstance(idx, str):
                idx = self.re.groupindex[idx]
        else:
            idx = 0
        return self.regs[idx]

    def start(self, *args) -> int:
        """
        start([group=0]) -> int.
        Return index of the start of the substring matched by group.
        """
        return self.span(*args)[0]

    def end(self, *args) -> int:
        """
        end([group=0]) -> int.
        Return index of the end of the substring matched by group.
        """
        return self.span(*args)[1]

    def _get_lastgroup(self):
        # No named capture groups? Return None
        if len(self.groupdict()) == 0:
            return None

        # Inverse map of dictionary
        inv_map = {value: key for key, value in
                self.re.groupindex.items()}
        
        # Last matched capturing group not in dictionary
        # -> not a named group.
        if self.lastindex not in inv_map.keys():
            return None

        # Return the name of the last named matched capturing group
        else:
            return inv_map[self.lastindex]



import pytest, re
import kre.kre as kre

arirang = "아리랑 아리랑 아라리요. 아리랑 고개로 넘어간다. 나를 버리고 가시는 님은 십리도 못가서 발병 난다."

news = '월간<산>은 무명명산들을 독자들의 도움을 통해 하나씩 찾아나간다. 국립, 군립, 도립공원 및 100대 명산에 해당하지 않는 산이면서 또 산행하는 것이 불법이 아닌 산이 대상이다. 직접 제보한 독자와 함께 오른다. 제보는 blackhouse@chosun.com 찬란했다. 분명히 빛나고 있었다. 햇빛이 들지 않는 깊은 산골인데도 눈이 부셨다. 구불거리는 계곡을 따라 흘러온 티 없이 투명한 물이 바위에 부딪쳐 맑게 부서지며 빛의 파편을 흩뿌린다. 그 광경에 눈이 멀 것 같아 잠시 이끼 낀 바위에 눈을 얹혀놓고 쉬어본다. 그러자 이번엔 계곡이 산의 정수리부터 꼭 잡아 끌고 내려온 냉기를 온 몸에 퍼붓는다. 살짝 솟았던 땀이 꼼짝없이 얼어붙는다. 귀를 앵앵거리는 날벌레도, 시끄러운 트로트 소리도 없는, 고요한 오지계곡의 호사다. 충북 제천 십자봉(983m)은 완전히 알려지지 않은 산은 아니다. 최근 등산을 시작한 이들에게는 깜깜할지 몰라도 약 30~40년 전부터 꾸준히 산악인들이 찾아온 바 있다. 십자봉을 제보해 준 독자 박노원씨는 "십자봉이 있는 제천시 백운면은 내 고향 시골"이라며 "현재 내 나이가 50대 후반인데 고등학교 무렵부터 산악인들이 찾기 시작했던 것으로 기억한다"고 전했다. 특히 발치에 흐르는 덕동계곡이 2000년대 이후 조용히 다녀오기 좋은 피서지로 유명세를 탔기에 이를 낀 십자봉의 존재도 덩달아 상기되곤 했다. 그래서 여름 무더위에 가기 딱 좋은 산으로 꼽힌다. 이처럼 이름값이 있지만 그래도 여전히 오지의 산임은 분명하다. 제천과 충주, 원주 사이에 있으며 서쪽에 미륵산, 북쪽에 백운산과 치악산에 둘러싸인 형세이기 때문에 개발의 손길이 거의 미치지 않았고, 등산객의 답압이 자연의 성장속도를 이기지 못해 등산로 대부분 수풀이 우거져 있다.'

nonsense = '할ㄱ으하느늘근ㅡ'

### Public Functions

def test_search():
    res = kre.search('ㅏㄴ[ㄱ-ㅣ ]+독', news)
    assert res.group() == "산들을 독"
    res = kre.search('ㅏㄴ[ㄱ-ㅣ >]+독', news)
    assert res.group() == "산>은 무명명산들을 독"
    res = kre.search('(ㄹ).*(ㄴ)', arirang)
    assert res.regs == ((1, 54), (1, 2), (53, 54))

def test_match():
    res = kre.match('ㅏㄴ[ㄱ-ㅣ ]+독', news)
    assert res == None
    res = kre.match('.ㅝ[^ㅎ]*', news)
    assert res.group() == "월간<산>은 무명명산들을 독자들의 도움을 통"
    assert res.span() == (0,24)
    assert res.start() == 0
    assert res.end() == 24
    res = kre.match(r"ㅎ", "한글") # 1st letter of 1st syllable
    assert res.span() == (0,1)
    res = kre.match(r"ㅏ", "한글") # 2nd letter of 1st syllable
    assert res == None

def test_fullmatch():
    assert kre.fullmatch('ㅏ','ㅏ').span() == (0,1)
    assert kre.fullmatch('ㅏ','ㅏㅏ') == None
    assert kre.fullmatch('ㅏ','가') == None

def test_sub():
    res = kre.sub('ㄴ다', 'ㄹ 거예요', arirang, syllabify="extended")
    assert res == "아리랑 아리랑 아라리요. 아리랑 고개로 넘어갈 거예요. 나를 버리고 가시는 님은 십리도 못가서 발병 날 거예요."

def test_sub_extended():
    # Ensure that extended will syllabify substituted portion with
    # preceding and following characters
    res = kre.sub('ㅡㄴ', 'ㅏㅂ', '하느늘ㅏ근', syllabify="extended")
    assert res == '하나브라갑'
    res = kre.sub('ㅡㄴ', 'ㅂ', '하느늘ㅏ근', syllabify="extended")
    assert res == '한브락ㅂ'
    # Ensure that it does not extended to two characters before
    res = kre.sub('ㅡㄴ', 'ㅂ', 'ㄱㅏ근ㅏ', syllabify="extended")
    assert res == 'ㄱㅏㄱ바'
    # Ensure that it does not extended to two characters after
    res = kre.sub('ㄱ', 'ㅂ', 'ㄱㅏㅁ', syllabify="extended")
    assert res == '바ㅁ'

def test_subn():
    assert kre.subn('a', 'b', 'c') == ('c', 0)

    res = kre.subn('느', '나가', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하나가나갈그나가', 3)

    res = kre.subn('ㅏ','ㅗ',nonsense, syllabify="extended")
    assert res == ('홁으호느늘근ㅡ', 2)

    res = kre.subn('ㅡ','ㅓ', nonsense, syllabify='none')
    assert res ==  ('할ㄱㅇㅓ하ㄴㅓㄴㅓㄹㄱㅓㄴㅓ', 5)

    # Test limiting substitutions with 'count' variable
    res = kre.subn('ㅡ','ㅓ', nonsense, count=2, syllabify='none')
    assert res ==  ('할ㄱㅇㅓ하ㄴㅓ늘근ㅡ', 2)
    
    # Substitution count != 'count' variable if fewer matches 
    res = kre.subn('ㅡ','ㅓ', nonsense, count=6, syllabify='none')
    assert res ==  ('할ㄱㅇㅓ하ㄴㅓㄴㅓㄹㄱㅓㄴㅓ', 5)

    res = kre.subn('ㅡ','ㅓ', nonsense, syllabify='minimal')
    assert res == ('할ㄱ어하너널건ㅓ', 5)

    res = kre.subn('ㅡ','ㅓ', nonsense, syllabify='extended')
    assert res == ('할ㄱ어하너널거너', 5)

    res = kre.subn('ㅡ','ㅓ', nonsense, syllabify='full')
    assert res == ('핡어하너널거너', 5)

    # Test multiple substitutions affecting same syllable
    res = kre.subn('ㅇ','ㄱ', '앙', syllabify="extended")
    assert res == ('각', 2) # former error: "강악"

    res = kre.subn('[ㅇ|ㅏ]','ㄱ', '앙', syllabify="extended")
    assert res == ('ㄱㄱㄱ', 3)

def test_empty_strings():
    res = kre.sub('a?', 'b', nonsense)
    assert res == ('bㅎbㅏbㄹbㄱbㅇbㅡbㅎbㅏbㄴbㅡbㄴbㅡbㄹbㄱbㅡbㄴbㅡb')

def test_split():
    pass

def test_findall():
    results = kre.findall('ㅏㄹ', arirang)
    assert results == ['아리', '아리', '아라', '라리', '아리', '나를', '발']

    res = kre.findall('ㄹ;', arirang, boundaries=True)
    assert res == ['를', '발']

    res = kre.findall(';ㄹ', arirang, boundaries=True)
    assert res == ['리', '랑', '리', '랑', '라', '리', '리', '랑', '로', '를', '리', '리']

    res = kre.findall('ㄹ.ㄹ', arirang)
    assert res == ['리랑', '리랑', '라리', '리랑', '를']

    res = kre.findall('ㄹ.ㄹ', arirang, boundaries=True)
    assert res == ['를']

    res = kre.findall('ㄹ;ㄹ', arirang, boundaries=True)
    assert res == None

    res = kre.findall('ㄹ.;ㄹ', arirang, boundaries=True)
    assert res == ['리랑', '리랑', '라리', '리랑']

    res = kre.findall(';ㄹ.ㄹ;', arirang, boundaries=True)
    assert res == ['를']

    res = kre.findall('ㄹ%', arirang, boundaries=True, delimiter='%')
    assert res == ['를', '발']


def test_finditer():
    pass

def test_compile():
    pass

def test_purge():
    pass

def test_escape():
    pass

### KRE_Pattern Method Calls
def test_kre_pattern_search():
    m = kre.compile('ㅏㄴ[ㄱ-ㅣ ]+독')
    res = m.search(news)
    assert res.group() == "산들을 독"
    m = kre.compile('ㅏㄴ[ㄱ-ㅣ >]+독')
    res = m.search(news)
    assert res.group() == "산>은 무명명산들을 독"

def test_kre_pattern_match():
    m = kre.compile('ㅏㄴ[ㄱ-ㅣ ]+독')
    res = m.match(news)
    assert res == None
    m = kre.compile('.ㅝ[^ㅎ]*')
    res = m.match(news)
    assert res.group() == "월간<산>은 무명명산들을 독자들의 도움을 통"
    assert res.span() == (0,24)
    assert res.start() == 0
    assert res.end() == 24

def test_kre_pattern_match_with_pos():
    p = kre.compile(r'ㅏ')
    assert p.match("한글") == None
    assert p.match("한글", 0).span() == (0,1)
    assert p.match("한글", 1) == None

def test_kre_pattern_match_with_pos_and_boundaries():
    # Test if pos expands to include preceding boundary
    # No boundaries in pattern
    p = kre.compile(r"ㄱ", boundaries=True)
    p.match("한글", 0) == None
    p.match("한글", 1).span() == (1,2)
    
    # Preceding boundary in pattern
    p = kre.compile(r";ㄱ", boundaries=True)
    p.match("한글", 1).span() == (1,2)
    
    # Following boundary in pattern
    p = kre.compile(r"ㄱ;", boundaries=True)
    p.match("한글", 1) == None
    p = kre.compile(r"ㄹ;", boundaries=True)
    p.match("한글", 1).span() == (1,2)

def test_kre_pattern_fullmatch():
    pass

def test_kre_pattern_fullmatch_with_pos_and_boundaries():
    p = kre.compile(r"ㅡㄹ", boundaries=True)
    p.match("한글", 1) == None
    p = kre.compile(r"ㅡㄹ;", boundaries=True)
    p.match("한글", 1).span == (1,2)


def test_kre_pattern_sub():
    m = kre.compile('ㄴ다')
    res = m.sub('ㄹ 거예요', arirang, syllabify="extended")
    assert res == "아리랑 아리랑 아라리요. 아리랑 고개로 넘어갈 거예요. 나를 버리고 가시는 님은 십리도 못가서 발병 날 거예요."

def test_kre_pattern_subn():
    m = kre.compile('느')
    res = m.subn('나가', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하나가나갈그나가', 3)

    m = kre.compile(';느', boundaries=True)
    res = m.subn(';나가', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하나가나갈근ㅡ', 2)

    # syllable-final boundary in pattern and sub
    m = kre.compile('[ㄱ|ㄴ];', boundaries=True)
    res = m.subn('ㅁ;',nonsense, syllabify="extended")
    assert res == ('할ㅁ으하느늘금ㅡ', 2)

    # syllable-final boundary in pattern but not sub
    m = kre.compile('[ㄱ|ㄴ];', boundaries=True)
    res = m.subn('ㅁ',nonsense, syllabify="extended")
    assert res == ('할ㅁ으하느늘그므', 2)

    m = kre.compile('ㅏ')
    res = m.subn('ㅗ',nonsense, syllabify="extended")
    assert res == ('홁으호느늘근ㅡ', 2)

    ### Boundaries True/False pairs
    m = kre.compile(r'ㅡ')
    res = m.subn('ㅓ', nonsense, syllabify='none')
    assert res ==  ('할ㄱㅇㅓ하ㄴㅓㄴㅓㄹㄱㅓㄴㅓ', 5)

    m = kre.compile('ㅡ', boundaries=True)
    res = m.subn('ㅓ', nonsense, syllabify='none')
    assert res ==  ('할ㄱㅇㅓ하ㄴㅓㄴㅓㄹㄱㅓㄴㅓ', 5)

    # syllabify='minimal': result may differ for patterns that cross
    # syllables
    # same
    res = m.subn('ㅓ', nonsense, syllabify='minimal')
    assert res == ('할ㄱ어하너널건ㅓ', 5)

    res = m.subn('ㅓ', nonsense, syllabify='minimal')
    assert res == ('할ㄱ어하너널건ㅓ', 5)

    # different
    m2 = kre.compile(r'ㅡㄴ')
    res = m2.subn(r'ㅓㄴ', nonsense, syllabify="minimal")
    assert res == ('할ㄱ으하너늘건ㅡ', 2)

    m2 = kre.compile(r'ㅡㄴ', boundaries=True)
    res = m2.subn(r'ㅓㄴ', nonsense, syllabify="minimal")
    assert res == ('할ㄱ으하느늘건ㅡ', 1)

    # syllabify='extended': result may differ for patterns that cross
    # syllables
    # same
    m = kre.compile(r'ㅡ')
    res = m.subn('ㅓ', nonsense, syllabify='extended')
    assert res == ('할ㄱ어하너널거너', 5)

    m = kre.compile(r'ㅡ', boundaries=True)
    res = m.subn('ㅓ', nonsense, syllabify='extended')
    assert res == ('할ㄱ어하너널건ㅓ', 5)

    # different
    m2 = kre.compile(r'ㅡㄴ')
    res = m2.subn(r'ㅓㄴ', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하너늘거느', 2)

    m2 = kre.compile(r'ㅡㄴ', boundaries=True)
    res = m2.subn(r'ㅓㄴ', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하느늘건ㅡ', 1)

    # syllabify='full'
    # same
    m = kre.compile(r'ㅡ')
    res = m.subn('ㅓ', nonsense, syllabify='full')
    assert res == ('핡어하너널거너', 5)

    m = kre.compile(r'ㅡ', boundaries=True)
    res = m.subn('ㅓ', nonsense, syllabify='full')
    assert res == ('핡어하너널거너', 5)

    # different
    m2 = kre.compile(r'ㅡㄴ')
    res = m2.subn('ㅓㄴ', nonsense, syllabify='full')
    assert res == ('핡으하너늘거느', 2)

    m2 = kre.compile(r'ㅡㄴ', boundaries=True)
    res = m2.subn('ㅓㄴ', nonsense, syllabify='full')
    assert res == ('핡으하느늘거느', 1)

    # With boundaries=True, 'minimal' and 'extended' differ only when
    # boundary from pattern is removed from repl argumenta
    m3 = kre.compile(r'ㅡㄴ;', boundaries=True)
    m4 = kre.compile(r';ㅡ', boundaries=True)
    
    res = m3.subn(r'ㅓㄴ', nonsense, syllabify="minimal")
    assert res == ('할ㄱ으하느늘건ㅡ', 1)

    res = m3.subn(r'ㅓㄴ', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하느늘거느', 1)

    res = m4.subn(r'ㅓ', nonsense, syllabify="minimal")
    assert res == ('할ㄱ으하느늘근ㅓ', 1)

    res = m4.subn(r'ㅓ', nonsense, syllabify="extended")
    assert res == ('할ㄱ으하느늘그너', 1)

def test_kre_pattern_split():
    pass

def test_kre_pattern_findall():
    m = kre.compile('ㅏㄹ')
    results = m.findall(arirang)
    assert results == ['아리', '아리', '아라', '라리', '아리', '나를', '발']

    m = kre.compile('ㄹ;', boundaries=True)
    res = m.findall(arirang)
    assert res == ['를', '발']

    m = kre.compile(';ㄹ', boundaries=True)
    res = m.findall(arirang)
    assert res == ['리', '랑', '리', '랑', '라', '리', '리', '랑', '로', '를', '리', '리']

    m = kre.compile('ㄹ.ㄹ')
    res = m.findall(arirang)
    assert res == ['리랑', '리랑', '라리', '리랑', '를']

    m = kre.compile('ㄹ.ㄹ', boundaries=True)
    res = m.findall(arirang)
    assert res == ['를']

    m = kre.compile('ㄹ;ㄹ', boundaries=True)
    res = m.findall(arirang)
    assert res == None

    m = kre.compile('ㄹ.;ㄹ', boundaries=True)
    res = m.findall(arirang)
    assert res == ['리랑', '리랑', '라리', '리랑']

    m = kre.compile(';ㄹ.ㄹ;', boundaries=True)
    res = m.findall(arirang)
    assert res == ['를']

    m = kre.compile('ㄹ%', boundaries=True, delimiter='%')
    res = m.findall(arirang)
    assert res == ['를', '발']

def test_kre_pattern_finditer():
    pass


### KRE_Match object attributes and functions

def test_kre_match_special_cases():
    # The following includes a named group, unmatched capturing group,
    # and empty string matching group.
    p = kre.compile(r"a(?P<f>f)(p)?(.*?)")
    m = p.search('sdaflkj')
    re_p = re.compile(r"a(?P<f>f)(p)?(.*?)")
    re_m = re_p.search('sdaflkj')
    assert m.regs == re_m.regs
    assert m.groups() == re_m.groups()
    for n in range(len(m.regs)):
        assert m.group(n) == re_m.group(n)


    # Test case: matches of empty strings, but at syllable boundaries
    # and within syllables. For such cases, re module returns span of
    # length 0 (pos==endpos), whereas kre returns length 1 when empty
    # string is syllable-internal.

def test_kre_match_object():
    # Test case: match of search is syllable, so result should be 
    # identical to standard re search

    # Pattern includes named and unnamed groups, as well as a named
    # group with 0 matches
    p = kre.compile('(?P<첫째>ㄱ)(으)(?P<h>h)?.*(?P<둘째>근)')
    m = p.search(nonsense)
    re_p = re.compile('(?P<첫째>ㄱ)(으)(?P<h>h)?.*(?P<둘째>근)')
    re_m = re_p.search(nonsense)

    # attributes
    assert m.pos == re_m.pos
    assert m.endpos == re_m.endpos
    # assert m.re == re_m.re # WILL CHANGE
    assert m.regs == re_m.regs
    assert m.string == re_m.string
    assert m.lastgroup == re_m.lastgroup
    assert m.lastindex == re_m.lastindex

    # methods
    assert m.end() == re_m.end()
    assert m.start() == re_m.start()
    assert m.groups() == re_m.groups()
    assert m.span() == re_m.span()
    # Change default value for unmatched groups (default is None)
    assert m.groups('changed') == re_m.groups('changed')
    for n in range(len(m.groups())+1):
        assert m.end(n) == re_m.end(n)
        assert m.start(n) == re_m.start(n)
        assert m.group(n) == re_m.group(n)
        assert m.span(n) == re_m.span(n)
    assert m.groupdict() == re_m.groupdict()
    # Change default value for unmatched named groups (default is None)
    assert m.groupdict('changed') == re_m.groupdict('changed')

    # not yet implemented
    #assert m.expand() == re_m.expand()

def test_kre_match_object2():
    # Test case: match of search is subsyllable, so result should be 
    # identical to standard re search

    # Pattern includes named and unnamed groups, as well as a named
    # group with 0 matches
    # kre searches for subsyllables should have same results as re
    # searches for syllables containing said subsyllables
    p = kre.compile('(?P<첫째>ㄱ).(ㅡ)(?P<h>h)?.*(?P<둘째>그)')
    m = p.search(nonsense)
    re_p = re.compile('(?P<첫째>ㄱ)(으)(?P<h>h)?.*(?P<둘째>근)')
    re_m = re_p.search(nonsense)

    # attributes
    assert m.pos == re_m.pos
    assert m.endpos == re_m.endpos
    # assert m.re == re_m.re # WILL CHANGE
    assert m.regs == re_m.regs
    assert m.string == re_m.string
    assert m.lastgroup == re_m.lastgroup
    assert m.lastindex == re_m.lastindex

    # methods
    assert m.end() == re_m.end()
    assert m.start() == re_m.start()
    assert m.groups() == re_m.groups()
    assert m.span() == re_m.span()
    # Change default value for unmatched groups (default is None)
    assert m.groups('changed') == re_m.groups('changed')
    for n in range(len(m.groups())+1):
        assert m.end(n) == re_m.end(n)
        assert m.start(n) == re_m.start(n)
        assert m.group(n) == re_m.group(n)
        assert m.span(n) == re_m.span(n)
    assert m.groupdict() == re_m.groupdict()
    # Change default value for unmatched named groups (default is None)
    assert m.groupdict('changed') == re_m.groupdict('changed')

    # not yet implemented
    #assert m.expand() == re_m.expand()

def test_kre_match_object3():
    # Test case: match of subsyllables pattern that crosses syllables. 
    # Result should be identical to standard re search off all involved
    # syllables

    # Pattern includes named and unnamed groups, as well as a named
    # group with 0 matches
    # kre searches for subsyllables should have same results as re
    # searches for syllables containing said subsyllables
    p = kre.compile('(?P<첫째>ㄱ.ㅡ)(?P<h>h)?.*(?P<둘째>그)')
    m = p.search(nonsense)
    re_p = re.compile('(?P<첫째>ㄱ으)(?P<h>h)?.*(?P<둘째>근)')
    re_m = re_p.search(nonsense)

    # attributes
    assert m.pos == re_m.pos
    assert m.endpos == re_m.endpos
    # assert m.re == re_m.re # WILL CHANGE
    assert m.regs == re_m.regs
    assert m.string == re_m.string
    assert m.lastgroup == re_m.lastgroup
    assert m.lastindex == re_m.lastindex

    # methods
    assert m.end() == re_m.end()
    assert m.start() == re_m.start()
    assert m.groups() == re_m.groups()
    assert m.span() == re_m.span()
    # Change default value for unmatched groups (default is None)
    assert m.groups('changed') == re_m.groups('changed')
    for n in range(len(m.groups())+1):
        assert m.end(n) == re_m.end(n)
        assert m.start(n) == re_m.start(n)
        assert m.group(n) == re_m.group(n)
        assert m.span(n) == re_m.span(n)
    assert m.groupdict() == re_m.groupdict()
    # Change default value for unmatched named groups (default is None)
    assert m.groupdict('changed') == re_m.groupdict('changed')

    # not yet implemented
    #assert m.expand() == re_m.expand()

def test_kre_match_object4():
    # Test case: match of subsyllables pattern that crosses syllables. 
    # Result should be identical to standard re search off all involved
    # syllables

    # Pattern includes named and unnamed groups, as well as a named
    # group with 0 matches
    # kre searches for subsyllables should have same results as re
    # searches for syllables containing said subsyllables
    p = kre.compile('(?P<첫째>ㄱ;.ㅡ)(?P<h>h)?.*(?P<둘째>그)',
            boundaries=True)
    m = p.search(nonsense)
    re_p = re.compile('(?P<첫째>ㄱ으)(?P<h>h)?.*(?P<둘째>근)')
    re_m = re_p.search(nonsense)

    # attributes
    assert m.pos == re_m.pos
    assert m.endpos == re_m.endpos
    # assert m.re == re_m.re # WILL CHANGE
    assert m.regs == re_m.regs
    assert m.string == re_m.string
    assert m.lastgroup == re_m.lastgroup
    assert m.lastindex == re_m.lastindex

    # methods
    assert m.end() == re_m.end()
    assert m.start() == re_m.start()
    assert m.groups() == re_m.groups()
    assert m.span() == re_m.span()
    # Change default value for unmatched groups (default is None)
    assert m.groups('changed') == re_m.groups('changed')
    for n in range(len(m.groups())+1):
        assert m.end(n) == re_m.end(n)
        assert m.start(n) == re_m.start(n)
        assert m.group(n) == re_m.group(n)
        assert m.span(n) == re_m.span(n)
    assert m.groupdict() == re_m.groupdict()
    # Change default value for unmatched named groups (default is None)
    assert m.groupdict('changed') == re_m.groupdict('changed')

    # not yet implemented
    #assert m.expand() == re_m.expand()

def test_kre_match_object5():
    # Test case: match of subsyllables pattern that crosses syllables. 
    # Result should be identical to standard re search off all involved
    # syllables

    # Pattern includes named and unnamed groups, as well as a named
    # group with 0 matches
    # kre searches for subsyllables should have same results as re
    # searches for syllables containing said subsyllables
    p = kre.compile('(?P<첫째>ㄱ)(.ㅡ).*(?P<둘째>그)(?P<h>h)?')
    m = p.search(nonsense)
    re_p = re.compile('(?P<첫째>ㄱ)(으).*(?P<둘째>근)(?P<h>h)?')
    re_m = re_p.search(nonsense)

    # attributes
    assert m.pos == re_m.pos
    assert m.endpos == re_m.endpos
    # assert m.re == re_m.re # WILL CHANGE
    assert m.regs == re_m.regs
    assert m.string == re_m.string
    assert m.lastgroup == re_m.lastgroup
    assert m.lastindex == re_m.lastindex

    # methods
    assert m.end() == re_m.end()
    assert m.start() == re_m.start()
    assert m.groups() == re_m.groups()
    assert m.span() == re_m.span()
    # Change default value for unmatched groups (default is None)
    assert m.groups('changed') == re_m.groups('changed')
    for n in range(len(m.groups())+1):
        assert m.end(n) == re_m.end(n)
        assert m.start(n) == re_m.start(n)
        assert m.group(n) == re_m.group(n)
        assert m.span(n) == re_m.span(n)
    assert m.groupdict() == re_m.groupdict()
    # Change default value for unmatched named groups (default is None)
    assert m.groupdict('changed') == re_m.groupdict('changed')

    # not yet implemented
    #assert m.expand() == re_m.expand()

### Mapping class tests 
def test_kre_mapping():
    ls = kre.Mapping('This is 한글ㅋㅋ.', boundaries=True)
    assert ls.delimited == 'This is ;한;글;ㅋ;ㅋ;.'
    assert ls.del2orig == (0, 1, 2, 3, 4, 5, 6, 7, None, 8, None, 9, 
            None, 10, None, 11, None, 12)


### KRE_Pattern Method Calls
